@@ -63,15 +63,16 @@ impl StatementGraph {
                 S::Switch {
                     selector,
                     ref cases,
-                    ref default,
                 } => {
                     self.dependencies.push((id, selector, "selector"));
                     for case in cases {
                         let case_id = self.add(&case.body);
-                        self.flow.push((id, case_id, "case"));
+                        let label = match case.value {
+                            crate::SwitchValue::Integer(_) => "case",
+                            crate::SwitchValue::Default => "default",
+                        };
+                        self.flow.push((id, case_id, label));
                     }
-                    let default_id = self.add(default);
-                    self.flow.push((id, default_id, "default"));
                     "Switch"
                 }
                 S::Loop {
@@ -230,6 +231,7 @@ fn write_fun(
             E::ImageSample {
                 image,
                 sampler,
+                gather,
                 coordinate,
                 array_index,
                 offset: _,
@@ -259,7 +261,11 @@ fn write_fun(
                 if let Some(expr) = depth_ref {
                     edges.insert("depth_ref", expr);
                 }
-                ("ImageSample".into(), 5)
+                let string = match gather {
+                    Some(component) => Cow::Owned(format!("ImageGather{:?}", component)),
+                    _ => Cow::Borrowed("ImageSample"),
+                };
+                (string, 5)
             }
             E::ImageLoad {
                 image,
