@@ -1,11 +1,8 @@
 //! Implementations for `BlockContext` methods.
 
 use super::{
-    index::{BoundsCheckResult, ExpressionPointer},
-    make_local,
-    selection::Selection,
-    Block, BlockContext, Dimension, Error, Instruction, LocalType, LookupType, LoopContext,
-    ResultMember, Writer, WriterFlags,
+    index::BoundsCheckResult, make_local, selection::Selection, Block, BlockContext, Dimension,
+    Error, Instruction, LocalType, LookupType, LoopContext, ResultMember, Writer, WriterFlags,
 };
 use crate::{arena::Handle, proc::TypeResolution};
 use spirv::Word;
@@ -19,6 +16,25 @@ fn get_dimension(type_inner: &crate::TypeInner) -> Dimension {
     }
 }
 
+/// The results of emitting code for a left-hand-side expression.
+///
+/// On success, `write_expression_pointer` returns one of these.
+enum ExpressionPointer {
+    /// The pointer to the expression's value is available, as the value of the
+    /// expression with the given id.
+    Ready { pointer_id: Word },
+
+    /// The access expression must be conditional on the value of `condition`, a boolean
+    /// expression that is true if all indices are in bounds. If `condition` is true, then
+    /// `access` is an `OpAccessChain` instruction that will compute a pointer to the
+    /// expression's value. If `condition` is false, then executing `access` would be
+    /// undefined behavior.
+    Conditional {
+        condition: Word,
+        access: Instruction,
+    },
+}
+
 impl Writer {
     // Flip Y coordinate to adjust for coordinate space difference
     // between SPIR-V and our IR.
@@ -210,6 +226,10 @@ impl<'w> BlockContext<'w> {
                     | crate::TypeInner::Matrix { .. }
                     | crate::TypeInner::Array { .. }
                     | crate::TypeInner::Struct { .. } => {
+                        // We never need bounds checks here: dynamically sized arrays can
+                        // only appear behind pointers, and are thus handled by the
+                        // `is_intermediate` case above. Everything else's size is
+                        // statically known and checked in validation.
                         let id = self.gen_id();
                         let base_id = self.cached[base];
                         block.body.push(Instruction::composite_extract(
@@ -227,7 +247,7 @@ impl<'w> BlockContext<'w> {
                 }
             }
             crate::Expression::GlobalVariable(handle) => {
-                self.writer.global_variables[handle.index()].id
+                self.writer.global_variables[handle.index()].access_id
             }
             crate::Expression::Constant(handle) => self.writer.constant_ids[handle.index()],
             crate::Expression::Splat { size, value } => {
@@ -346,6 +366,10 @@ impl<'w> BlockContext<'w> {
                         (Dimension::Matrix, Dimension::Scalar { .. }) => {
                             spirv::Op::MatrixTimesScalar
                         }
+                        (Dimension::Scalar, Dimension::Matrix { .. }) => {
+                            preserve_order = false;
+                            spirv::Op::MatrixTimesScalar
+                        }
                         (Dimension::Matrix, Dimension::Vector) => spirv::Op::MatrixTimesVector,
                         (Dimension::Matrix, Dimension::Matrix) => spirv::Op::MatrixTimesMatrix,
                         (Dimension::Vector, Dimension::Vector)
@@ -356,7 +380,6 @@ impl<'w> BlockContext<'w> {
                         }
                         (Dimension::Vector, Dimension::Vector)
                         | (Dimension::Scalar, Dimension::Scalar) => spirv::Op::IMul,
-                        other => unimplemented!("Mul {:?}", other),
                     },
                     crate::BinaryOperator::Divide => match left_ty_inner.scalar_kind() {
                         Some(crate::ScalarKind::Sint) => spirv::Op::SDiv,
@@ -410,9 +433,15 @@ impl<'w> BlockContext<'w> {
                         Some(crate::ScalarKind::Float) => spirv::Op::FOrdGreaterThanEqual,
                         _ => unimplemented!(),
                     },
-                    crate::BinaryOperator::And => spirv::Op::BitwiseAnd,
+                    crate::BinaryOperator::And => match left_ty_inner.scalar_kind() {
+                        Some(crate::ScalarKind::Bool) => spirv::Op::LogicalAnd,
+                        _ => spirv::Op::BitwiseAnd,
+                    },
                     crate::BinaryOperator::ExclusiveOr => spirv::Op::BitwiseXor,
-                    crate::BinaryOperator::InclusiveOr => spirv::Op::BitwiseOr,
+                    crate::BinaryOperator::InclusiveOr => match left_ty_inner.scalar_kind() {
+                        Some(crate::ScalarKind::Bool) => spirv::Op::LogicalOr,
+                        _ => spirv::Op::BitwiseOr,
+                    },
                     crate::BinaryOperator::LogicalAnd => spirv::Op::LogicalAnd,
                     crate::BinaryOperator::LogicalOr => spirv::Op::LogicalOr,
                     crate::BinaryOperator::ShiftLeft => spirv::Op::ShiftLeftLogical,
@@ -862,6 +891,7 @@ impl<'w> BlockContext<'w> {
             crate::Expression::ImageSample {
                 image,
                 sampler,
+                gather,
                 coordinate,
                 array_index,
                 offset,
@@ -871,6 +901,7 @@ impl<'w> BlockContext<'w> {
                 result_type_id,
                 image,
                 sampler,
+                gather,
                 coordinate,
                 array_index,
                 offset,
@@ -970,7 +1001,7 @@ impl<'w> BlockContext<'w> {
         Ok(())
     }
 
-    /// Build an `OpAccessChain` instruction for a left-hand-side expression.
+    /// Build an `OpAccessChain` instruction.
     ///
     /// Emit any needed bounds-checking expressions to `block`.
     ///
@@ -1042,7 +1073,7 @@ impl<'w> BlockContext<'w> {
                 }
                 crate::Expression::GlobalVariable(handle) => {
                     let gv = &self.writer.global_variables[handle.index()];
-                    break gv.id;
+                    break gv.access_id;
                 }
                 crate::Expression::LocalVariable(variable) => {
                     let local_var = &self.function.variables[&variable];
@@ -1151,7 +1182,6 @@ impl<'w> BlockContext<'w> {
                 crate::Statement::Switch {
                     selector,
                     ref cases,
-                    ref default,
                 } => {
                     let selector_id = self.cached[selector];
 
@@ -1162,13 +1192,30 @@ impl<'w> BlockContext<'w> {
                     ));
 
                     let default_id = self.gen_id();
-                    let raw_cases = cases
-                        .iter()
-                        .map(|c| super::instructions::Case {
-                            value: c.value as Word,
-                            label_id: self.gen_id(),
-                        })
-                        .collect::<Vec<_>>();
+
+                    let mut reached_default = false;
+                    let mut raw_cases = Vec::with_capacity(cases.len());
+                    let mut case_ids = Vec::with_capacity(cases.len());
+                    for case in cases.iter() {
+                        match case.value {
+                            crate::SwitchValue::Integer(value) => {
+                                let label_id = self.gen_id();
+                                // No cases should be added after the default case is encountered
+                                // since the default case catches all
+                                if !reached_default {
+                                    raw_cases.push(super::instructions::Case {
+                                        value: value as Word,
+                                        label_id,
+                                    });
+                                }
+                                case_ids.push(label_id);
+                            }
+                            crate::SwitchValue::Default => {
+                                case_ids.push(default_id);
+                                reached_default = true;
+                            }
+                        }
+                    }
 
                     self.function.consume(
                         block,
@@ -1180,24 +1227,25 @@ impl<'w> BlockContext<'w> {
                         ..loop_context
                     };
 
-                    for (i, (case, raw_case)) in cases.iter().zip(raw_cases.iter()).enumerate() {
+                    for (i, (case, label_id)) in cases.iter().zip(case_ids.iter()).enumerate() {
                         let case_finish_id = if case.fall_through {
-                            match raw_cases.get(i + 1) {
-                                Some(rc) => rc.label_id,
-                                None => default_id,
-                            }
+                            case_ids[i + 1]
                         } else {
                             merge_id
                         };
                         self.write_block(
-                            raw_case.label_id,
+                            *label_id,
                             &case.body,
                             Some(case_finish_id),
                             inner_context,
                         )?;
                     }
 
-                    self.write_block(default_id, default, Some(merge_id), inner_context)?;
+                    // If no default was encountered write a empty block to satisfy the presence of
+                    // a block the default label
+                    if !reached_default {
+                        self.write_block(default_id, &[], Some(merge_id), inner_context)?;
+                    }
 
                     block = Block::new(merge_id);
                 }
