@@ -54,294 +54,496 @@ impl Parser {
 
         match fc {
             FunctionCallKind::TypeConstructor(ty) => {
-                let h = if args.len() == 1 {
-                    let expr_type = self.resolve_type(ctx, args[0].0, args[0].1)?;
+                if args.len() == 1 {
+                    self.constructor_single(ctx, body, ty, args[0], meta)
+                        .map(Some)
+                } else {
+                    self.constructor_many(ctx, body, ty, args, meta).map(Some)
+                }
+            }
+            FunctionCallKind::Function(name) => {
+                self.function_call(ctx, stmt, body, name, args, raw_args, meta)
+            }
+        }
+    }
 
-                    let vector_size = match *expr_type {
-                        TypeInner::Vector { size, .. } => Some(size),
-                        _ => None,
-                    };
+    fn constructor_single(
+        &mut self,
+        ctx: &mut Context,
+        body: &mut Block,
+        ty: Handle<Type>,
+        (mut value, expr_meta): (Handle<Expression>, Span),
+        meta: Span,
+    ) -> Result<Handle<Expression>> {
+        let expr_type = self.resolve_type(ctx, value, expr_meta)?;
 
-                    // Special case: if casting from a bool, we need to use Select and not As.
-                    match self.module.types[ty].inner.scalar_kind() {
-                        Some(result_scalar_kind)
-                            if expr_type.scalar_kind() == Some(ScalarKind::Bool)
-                                && result_scalar_kind != ScalarKind::Bool =>
-                        {
-                            let (condition, expr_meta) = args[0];
-                            let c0 = self.add_constant_value(result_scalar_kind, 0u64, meta);
-                            let c1 = self.add_constant_value(result_scalar_kind, 1u64, meta);
-                            let mut reject =
-                                ctx.add_expression(Expression::Constant(c0), expr_meta, body);
-                            let mut accept =
-                                ctx.add_expression(Expression::Constant(c1), expr_meta, body);
+        let vector_size = match *expr_type {
+            TypeInner::Vector { size, .. } => Some(size),
+            _ => None,
+        };
 
-                            ctx.implicit_splat(self, &mut reject, meta, vector_size)?;
-                            ctx.implicit_splat(self, &mut accept, meta, vector_size)?;
+        // Special case: if casting from a bool, we need to use Select and not As.
+        match self.module.types[ty].inner.scalar_kind() {
+            Some(result_scalar_kind)
+                if expr_type.scalar_kind() == Some(ScalarKind::Bool)
+                    && result_scalar_kind != ScalarKind::Bool =>
+            {
+                let c0 = self.add_constant_value(result_scalar_kind, 0u64, meta);
+                let c1 = self.add_constant_value(result_scalar_kind, 1u64, meta);
+                let mut reject = ctx.add_expression(Expression::Constant(c0), expr_meta, body);
+                let mut accept = ctx.add_expression(Expression::Constant(c1), expr_meta, body);
 
-                            let h = ctx.add_expression(
-                                Expression::Select {
-                                    accept,
-                                    reject,
-                                    condition,
-                                },
-                                expr_meta,
-                                body,
-                            );
+                ctx.implicit_splat(self, &mut reject, meta, vector_size)?;
+                ctx.implicit_splat(self, &mut accept, meta, vector_size)?;
 
-                            return Ok(Some(h));
-                        }
-                        _ => {}
-                    }
+                let h = ctx.add_expression(
+                    Expression::Select {
+                        accept,
+                        reject,
+                        condition: value,
+                    },
+                    expr_meta,
+                    body,
+                );
 
-                    match self.module.types[ty].inner {
-                        TypeInner::Vector { size, kind, width } if vector_size.is_none() => {
-                            let (mut value, meta) = args[0];
-                            ctx.implicit_conversion(self, &mut value, meta, kind, width)?;
+                return Ok(h);
+            }
+            _ => {}
+        }
 
-                            ctx.add_expression(Expression::Splat { size, value }, meta, body)
-                        }
-                        TypeInner::Scalar { kind, width } => ctx.add_expression(
-                            Expression::As {
-                                kind,
-                                expr: args[0].0,
-                                convert: Some(width),
-                            },
-                            args[0].1,
-                            body,
-                        ),
-                        TypeInner::Vector { size, kind, width } => {
-                            let mut expr = args[0].0;
+        Ok(match self.module.types[ty].inner {
+            TypeInner::Vector { size, kind, width } if vector_size.is_none() => {
+                ctx.implicit_conversion(self, &mut value, meta, kind, width)?;
 
-                            if vector_size.map_or(true, |s| s != size) {
-                                expr = ctx.vector_resize(size, expr, args[0].1, body);
-                            }
+                if let TypeInner::Scalar { .. } = *self.resolve_type(ctx, value, expr_meta)? {
+                    ctx.add_expression(Expression::Splat { size, value }, meta, body)
+                } else {
+                    self.vector_constructor(
+                        ctx,
+                        body,
+                        ty,
+                        size,
+                        kind,
+                        width,
+                        &[(value, expr_meta)],
+                        meta,
+                    )?
+                }
+            }
+            TypeInner::Scalar { kind, width } => {
+                let mut expr = value;
+                if let TypeInner::Vector { .. } | TypeInner::Matrix { .. } =
+                    *self.resolve_type(ctx, value, expr_meta)?
+                {
+                    expr = ctx.add_expression(
+                        Expression::AccessIndex {
+                            base: expr,
+                            index: 0,
+                        },
+                        meta,
+                        body,
+                    );
+                }
 
-                            ctx.add_expression(
-                                Expression::As {
-                                    kind,
-                                    expr,
-                                    convert: Some(width),
-                                },
-                                args[0].1,
-                                body,
-                            )
-                        }
-                        TypeInner::Matrix {
-                            columns,
-                            rows,
+                if let TypeInner::Matrix { .. } = *self.resolve_type(ctx, value, expr_meta)? {
+                    expr = ctx.add_expression(
+                        Expression::AccessIndex {
+                            base: expr,
+                            index: 0,
+                        },
+                        meta,
+                        body,
+                    );
+                }
+
+                ctx.add_expression(
+                    Expression::As {
+                        kind,
+                        expr,
+                        convert: Some(width),
+                    },
+                    meta,
+                    body,
+                )
+            }
+            TypeInner::Vector { size, kind, width } => {
+                if vector_size.map_or(true, |s| s != size) {
+                    value = ctx.vector_resize(size, value, expr_meta, body);
+                }
+
+                ctx.add_expression(
+                    Expression::As {
+                        kind,
+                        expr: value,
+                        convert: Some(width),
+                    },
+                    meta,
+                    body,
+                )
+            }
+            TypeInner::Matrix {
+                columns,
+                rows,
+                width,
+            } => self.matrix_one_arg(
+                ctx,
+                body,
+                ty,
+                columns,
+                rows,
+                width,
+                (value, expr_meta),
+                meta,
+            )?,
+            TypeInner::Struct { .. } | TypeInner::Array { .. } => ctx.add_expression(
+                Expression::Compose {
+                    ty,
+                    components: vec![value],
+                },
+                meta,
+                body,
+            ),
+            _ => {
+                self.errors.push(Error {
+                    kind: ErrorKind::SemanticError("Bad type constructor".into()),
+                    meta,
+                });
+
+                value
+            }
+        })
+    }
+
+    #[allow(clippy::too_many_arguments)]
+    fn matrix_one_arg(
+        &mut self,
+        ctx: &mut Context,
+        body: &mut Block,
+        ty: Handle<Type>,
+        columns: crate::VectorSize,
+        rows: crate::VectorSize,
+        width: crate::Bytes,
+        (mut value, expr_meta): (Handle<Expression>, Span),
+        meta: Span,
+    ) -> Result<Handle<Expression>> {
+        let mut components = Vec::with_capacity(columns as usize);
+        // TODO: casts
+        // `Expression::As` doesn't support matrix width
+        // casts so we need to do some extra work for casts
+
+        ctx.implicit_conversion(self, &mut value, expr_meta, ScalarKind::Float, width)?;
+        match *self.resolve_type(ctx, value, expr_meta)? {
+            TypeInner::Scalar { .. } => {
+                // If a matrix is constructed with a single scalar value, then that
+                // value is used to initialize all the values along the diagonal of
+                // the matrix; the rest are given zeros.
+                let vector_ty = self.module.types.insert(
+                    Type {
+                        name: None,
+                        inner: TypeInner::Vector {
+                            size: rows,
+                            kind: ScalarKind::Float,
                             width,
-                        } => {
-                            // TODO: casts
-                            // `Expression::As` doesn't support matrix width
-                            // casts so we need to do some extra work for casts
+                        },
+                    },
+                    meta,
+                );
+                let zero_constant = self.module.constants.fetch_or_append(
+                    Constant {
+                        name: None,
+                        specialization: None,
+                        inner: ConstantInner::Scalar {
+                            width,
+                            value: ScalarValue::Float(0.0),
+                        },
+                    },
+                    meta,
+                );
+                let zero = ctx.add_expression(Expression::Constant(zero_constant), meta, body);
 
-                            let (mut value, meta) = args[0];
-                            ctx.implicit_conversion(
-                                self,
-                                &mut value,
-                                meta,
-                                ScalarKind::Float,
-                                width,
-                            )?;
-                            match *self.resolve_type(ctx, value, meta)? {
-                                TypeInner::Scalar { .. } => {
-                                    // If a matrix is constructed with a single scalar value, then that
-                                    // value is used to initialize all the values along the diagonal of
-                                    // the matrix; the rest are given zeros.
-                                    let mut components = Vec::with_capacity(columns as usize);
-                                    let vector_ty = self.module.types.insert(
-                                        Type {
-                                            name: None,
-                                            inner: TypeInner::Vector {
-                                                size: rows,
-                                                kind: ScalarKind::Float,
-                                                width,
-                                            },
-                                        },
-                                        meta,
-                                    );
-                                    let zero_constant = self.module.constants.fetch_or_append(
-                                        Constant {
-                                            name: None,
-                                            specialization: None,
-                                            inner: ConstantInner::Scalar {
-                                                width,
-                                                value: ScalarValue::Float(0.0),
-                                            },
-                                        },
-                                        meta,
-                                    );
-                                    let zero = ctx.add_expression(
-                                        Expression::Constant(zero_constant),
-                                        meta,
-                                        body,
-                                    );
-
-                                    for i in 0..columns as u32 {
-                                        components.push(
-                                            ctx.add_expression(
-                                                Expression::Compose {
-                                                    ty: vector_ty,
-                                                    components: (0..rows as u32)
-                                                        .into_iter()
-                                                        .map(|r| match r == i {
-                                                            true => value,
-                                                            false => zero,
-                                                        })
-                                                        .collect(),
-                                                },
-                                                meta,
-                                                body,
-                                            ),
-                                        )
-                                    }
-
-                                    ctx.add_expression(
-                                        Expression::Compose { ty, components },
-                                        meta,
-                                        body,
-                                    )
-                                }
-                                TypeInner::Matrix { rows: ori_rows, .. } => {
-                                    let mut components = Vec::new();
-
-                                    for n in 0..columns as u32 {
-                                        let mut vector = ctx.add_expression(
-                                            Expression::AccessIndex {
-                                                base: value,
-                                                index: n,
-                                            },
-                                            meta,
-                                            body,
-                                        );
-
-                                        if ori_rows != rows {
-                                            vector = ctx.vector_resize(rows, vector, meta, body);
-                                        }
-
-                                        components.push(vector)
-                                    }
-
-                                    ctx.add_expression(
-                                        Expression::Compose { ty, components },
-                                        meta,
-                                        body,
-                                    )
-                                }
-                                _ => {
-                                    let columns =
-                                        iter::repeat(value).take(columns as usize).collect();
-
-                                    ctx.add_expression(
-                                        Expression::Compose {
-                                            ty,
-                                            components: columns,
-                                        },
-                                        meta,
-                                        body,
-                                    )
-                                }
-                            }
-                        }
-                        TypeInner::Struct { .. } | TypeInner::Array { .. } => ctx.add_expression(
+                for i in 0..columns as u32 {
+                    components.push(
+                        ctx.add_expression(
                             Expression::Compose {
-                                ty,
-                                components: args.into_iter().map(|arg| arg.0).collect(),
+                                ty: vector_ty,
+                                components: (0..rows as u32)
+                                    .into_iter()
+                                    .map(|r| match r == i {
+                                        true => value,
+                                        false => zero,
+                                    })
+                                    .collect(),
                             },
                             meta,
                             body,
                         ),
-                        _ => {
-                            self.errors.push(Error {
-                                kind: ErrorKind::SemanticError("Bad type constructor".into()),
-                                meta,
-                            });
+                    )
+                }
+            }
+            TypeInner::Matrix {
+                rows: ori_rows,
+                columns: ori_cols,
+                ..
+            } => {
+                // If a matrix is constructed from a matrix, then each component
+                // (column i, row j) in the result that has a corresponding component
+                // (column i, row j) in the argument will be initialized from there. All
+                // other components will be initialized to the identity matrix.
 
-                            args[0].0
-                        }
-                    }
-                } else {
-                    let mut components = Vec::with_capacity(args.len());
-
-                    match self.module.types[ty].inner {
-                        TypeInner::Matrix {
-                            columns,
-                            rows,
+                let zero_constant = self.module.constants.fetch_or_append(
+                    Constant {
+                        name: None,
+                        specialization: None,
+                        inner: ConstantInner::Scalar {
                             width,
-                        } => {
-                            let mut flattened =
-                                Vec::with_capacity(columns as usize * rows as usize);
+                            value: ScalarValue::Float(0.0),
+                        },
+                    },
+                    meta,
+                );
+                let zero = ctx.add_expression(Expression::Constant(zero_constant), meta, body);
+                let one_constant = self.module.constants.fetch_or_append(
+                    Constant {
+                        name: None,
+                        specialization: None,
+                        inner: ConstantInner::Scalar {
+                            width,
+                            value: ScalarValue::Float(1.0),
+                        },
+                    },
+                    meta,
+                );
+                let one = ctx.add_expression(Expression::Constant(one_constant), meta, body);
+                let vector_ty = self.module.types.insert(
+                    Type {
+                        name: None,
+                        inner: TypeInner::Vector {
+                            size: rows,
+                            kind: ScalarKind::Float,
+                            width,
+                        },
+                    },
+                    meta,
+                );
 
-                            for (mut arg, meta) in args.iter().copied() {
-                                let scalar_components =
-                                    scalar_components(&self.module.types[ty].inner);
-                                if let Some((kind, width)) = scalar_components {
-                                    ctx.implicit_conversion(self, &mut arg, meta, kind, width)?;
-                                }
+                for i in 0..columns as u32 {
+                    if i < ori_cols as u32 {
+                        use std::cmp::Ordering;
 
-                                match *self.resolve_type(ctx, arg, meta)? {
-                                    TypeInner::Vector { size, .. } => {
-                                        for i in 0..(size as u32) {
-                                            flattened.push(ctx.add_expression(
+                        let vector = ctx.add_expression(
+                            Expression::AccessIndex {
+                                base: value,
+                                index: i,
+                            },
+                            meta,
+                            body,
+                        );
+
+                        components.push(match ori_rows.cmp(&rows) {
+                            Ordering::Less => {
+                                let components = (0..rows as u32)
+                                    .into_iter()
+                                    .map(|r| {
+                                        if r < ori_rows as u32 {
+                                            ctx.add_expression(
                                                 Expression::AccessIndex {
-                                                    base: arg,
-                                                    index: i,
+                                                    base: vector,
+                                                    index: r,
                                                 },
                                                 meta,
                                                 body,
-                                            ))
+                                            )
+                                        } else if r == i {
+                                            one
+                                        } else {
+                                            zero
                                         }
-                                    }
-                                    _ => flattened.push(arg),
-                                }
-                            }
+                                    })
+                                    .collect();
 
-                            let ty = self.module.types.insert(
-                                Type {
-                                    name: None,
-                                    inner: TypeInner::Vector {
-                                        size: rows,
-                                        kind: ScalarKind::Float,
-                                        width,
-                                    },
-                                },
-                                meta,
-                            );
-
-                            for chunk in flattened.chunks(rows as usize) {
-                                components.push(ctx.add_expression(
+                                ctx.add_expression(
                                     Expression::Compose {
-                                        ty,
-                                        components: Vec::from(chunk),
+                                        ty: vector_ty,
+                                        components,
+                                    },
+                                    meta,
+                                    body,
+                                )
+                            }
+                            Ordering::Equal => vector,
+                            Ordering::Greater => ctx.vector_resize(rows, vector, meta, body),
+                        })
+                    } else {
+                        let vec_constant = self.module.constants.fetch_or_append(
+                            Constant {
+                                name: None,
+                                specialization: None,
+                                inner: ConstantInner::Composite {
+                                    ty: vector_ty,
+                                    components: (0..rows as u32)
+                                        .into_iter()
+                                        .map(|r| match r == i {
+                                            true => one_constant,
+                                            false => zero_constant,
+                                        })
+                                        .collect(),
+                                },
+                            },
+                            meta,
+                        );
+                        let vec =
+                            ctx.add_expression(Expression::Constant(vec_constant), meta, body);
+
+                        components.push(vec)
+                    }
+                }
+            }
+            _ => {
+                components = iter::repeat(value).take(columns as usize).collect();
+            }
+        }
+
+        Ok(ctx.add_expression(Expression::Compose { ty, components }, meta, body))
+    }
+
+    #[allow(clippy::too_many_arguments)]
+    fn vector_constructor(
+        &mut self,
+        ctx: &mut Context,
+        body: &mut Block,
+        ty: Handle<Type>,
+        size: crate::VectorSize,
+        kind: ScalarKind,
+        width: crate::Bytes,
+        args: &[(Handle<Expression>, Span)],
+        meta: Span,
+    ) -> Result<Handle<Expression>> {
+        let mut components = Vec::with_capacity(size as usize);
+
+        for (mut arg, expr_meta) in args.iter().copied() {
+            ctx.implicit_conversion(self, &mut arg, expr_meta, kind, width)?;
+
+            if components.len() >= size as usize {
+                break;
+            }
+
+            match *self.resolve_type(ctx, arg, expr_meta)? {
+                TypeInner::Scalar { .. } => components.push(arg),
+                TypeInner::Matrix { rows, columns, .. } => {
+                    components.reserve(rows as usize * columns as usize);
+                    for c in 0..(columns as u32) {
+                        let base = ctx.add_expression(
+                            Expression::AccessIndex {
+                                base: arg,
+                                index: c,
+                            },
+                            expr_meta,
+                            body,
+                        );
+                        for r in 0..(rows as u32) {
+                            components.push(ctx.add_expression(
+                                Expression::AccessIndex { base, index: r },
+                                expr_meta,
+                                body,
+                            ))
+                        }
+                    }
+                }
+                TypeInner::Vector { size: ori_size, .. } => {
+                    components.reserve(ori_size as usize);
+                    for index in 0..(ori_size as u32) {
+                        components.push(ctx.add_expression(
+                            Expression::AccessIndex { base: arg, index },
+                            expr_meta,
+                            body,
+                        ))
+                    }
+                }
+                _ => components.push(arg),
+            }
+        }
+
+        components.truncate(size as usize);
+
+        Ok(ctx.add_expression(Expression::Compose { ty, components }, meta, body))
+    }
+
+    fn constructor_many(
+        &mut self,
+        ctx: &mut Context,
+        body: &mut Block,
+        ty: Handle<Type>,
+        args: Vec<(Handle<Expression>, Span)>,
+        meta: Span,
+    ) -> Result<Handle<Expression>> {
+        let mut components = Vec::with_capacity(args.len());
+
+        match self.module.types[ty].inner {
+            TypeInner::Matrix {
+                columns,
+                rows,
+                width,
+            } => {
+                let mut flattened = Vec::with_capacity(columns as usize * rows as usize);
+
+                for (mut arg, meta) in args.iter().copied() {
+                    let scalar_components = scalar_components(&self.module.types[ty].inner);
+                    if let Some((kind, width)) = scalar_components {
+                        ctx.implicit_conversion(self, &mut arg, meta, kind, width)?;
+                    }
+
+                    match *self.resolve_type(ctx, arg, meta)? {
+                        TypeInner::Vector { size, .. } => {
+                            for i in 0..(size as u32) {
+                                flattened.push(ctx.add_expression(
+                                    Expression::AccessIndex {
+                                        base: arg,
+                                        index: i,
                                     },
                                     meta,
                                     body,
                                 ))
                             }
                         }
-                        _ => {
-                            for (mut arg, meta) in args.iter().copied() {
-                                let scalar_components =
-                                    scalar_components(&self.module.types[ty].inner);
-                                if let Some((kind, width)) = scalar_components {
-                                    ctx.implicit_conversion(self, &mut arg, meta, kind, width)?;
-                                }
-
-                                components.push(arg)
-                            }
-                        }
+                        _ => flattened.push(arg),
                     }
+                }
 
-                    ctx.add_expression(Expression::Compose { ty, components }, meta, body)
-                };
+                let ty = self.module.types.insert(
+                    Type {
+                        name: None,
+                        inner: TypeInner::Vector {
+                            size: rows,
+                            kind: ScalarKind::Float,
+                            width,
+                        },
+                    },
+                    meta,
+                );
 
-                Ok(Some(h))
+                for chunk in flattened.chunks(rows as usize) {
+                    components.push(ctx.add_expression(
+                        Expression::Compose {
+                            ty,
+                            components: Vec::from(chunk),
+                        },
+                        meta,
+                        body,
+                    ))
+                }
             }
-            FunctionCallKind::Function(name) => {
-                self.function_call(ctx, stmt, body, name, args, raw_args, meta)
+            TypeInner::Vector { size, kind, width } => {
+                return self.vector_constructor(ctx, body, ty, size, kind, width, &args, meta)
+            }
+            _ => {
+                return Err(Error {
+                    kind: ErrorKind::SemanticError("Constructor: Too many arguments".into()),
+                    meta,
+                })
             }
         }
+
+        Ok(ctx.add_expression(Expression::Compose { ty, components }, meta, body))
     }
 
     #[allow(clippy::too_many_arguments)]
@@ -856,6 +1058,82 @@ impl Parser {
         });
     }
 
+    /// Helper function for building the input/output interface of the entry point
+    ///
+    /// Calls `f` with the data of the entry point argument, flattening composite types
+    /// recursively
+    ///
+    /// The passed arguments to the callback are:
+    /// - The name
+    /// - The pointer expression to the global storage
+    /// - The handle to the type of the entry point argument
+    /// - The binding of the entry point argument
+    /// - The expression arena
+    fn arg_type_walker(
+        &self,
+        name: Option<String>,
+        binding: crate::Binding,
+        pointer: Handle<Expression>,
+        ty: Handle<Type>,
+        expressions: &mut Arena<Expression>,
+        f: &mut impl FnMut(
+            Option<String>,
+            Handle<Expression>,
+            Handle<Type>,
+            crate::Binding,
+            &mut Arena<Expression>,
+        ),
+    ) {
+        match self.module.types[ty].inner {
+            TypeInner::Struct { ref members, .. } => {
+                let mut location = match binding {
+                    crate::Binding::Location { location, .. } => location,
+                    _ => return,
+                };
+
+                for (i, member) in members.iter().enumerate() {
+                    let member_pointer = expressions.append(
+                        Expression::AccessIndex {
+                            base: pointer,
+                            index: i as u32,
+                        },
+                        crate::Span::default(),
+                    );
+
+                    let binding = match member.binding.clone() {
+                        Some(binding) => binding,
+                        None => {
+                            let interpolation = self.module.types[member.ty]
+                                .inner
+                                .scalar_kind()
+                                .map(|kind| match kind {
+                                    ScalarKind::Float => crate::Interpolation::Perspective,
+                                    _ => crate::Interpolation::Flat,
+                                });
+                            let binding = crate::Binding::Location {
+                                location,
+                                interpolation,
+                                sampling: None,
+                            };
+                            location += 1;
+                            binding
+                        }
+                    };
+
+                    self.arg_type_walker(
+                        member.name.clone(),
+                        binding,
+                        member_pointer,
+                        member.ty,
+                        expressions,
+                        f,
+                    )
+                }
+            }
+            _ => f(name, pointer, ty, binding, expressions),
+        }
+    }
+
     pub(crate) fn add_entry_point(
         &mut self,
         function: Handle<Function>,
@@ -877,20 +1155,29 @@ impl Parser {
                 continue;
             }
 
-            let ty = self.module.global_variables[arg.handle].ty;
-            let idx = arguments.len() as u32;
-
-            arguments.push(FunctionArgument {
-                name: arg.name.clone(),
-                ty,
-                binding: Some(arg.binding.clone()),
-            });
-
             let pointer =
                 expressions.append(Expression::GlobalVariable(arg.handle), Default::default());
-            let value = expressions.append(Expression::FunctionArgument(idx), Default::default());
 
-            body.push(Statement::Store { pointer, value }, Default::default());
+            self.arg_type_walker(
+                arg.name.clone(),
+                arg.binding.clone(),
+                pointer,
+                self.module.global_variables[arg.handle].ty,
+                &mut expressions,
+                &mut |name, pointer, ty, binding, expressions| {
+                    let idx = arguments.len() as u32;
+
+                    arguments.push(FunctionArgument {
+                        name,
+                        ty,
+                        binding: Some(binding),
+                    });
+
+                    let value =
+                        expressions.append(Expression::FunctionArgument(idx), Default::default());
+                    body.push(Statement::Store { pointer, value }, Default::default());
+                },
+            )
         }
 
         body.extend_block(global_init_body);
@@ -913,37 +1200,41 @@ impl Parser {
                 continue;
             }
 
-            let ty = self.module.global_variables[arg.handle].ty;
-
-            members.push(StructMember {
-                name: arg.name.clone(),
-                ty,
-                binding: Some(arg.binding.clone()),
-                offset: span,
-            });
-
-            span += self.module.types[ty].inner.span(&self.module.constants);
-
             let pointer =
                 expressions.append(Expression::GlobalVariable(arg.handle), Default::default());
-            let len = expressions.len();
-            let load = expressions.append(Expression::Load { pointer }, Default::default());
-            body.push(
-                Statement::Emit(expressions.range_from(len)),
-                Default::default(),
-            );
-            components.push(load)
+
+            self.arg_type_walker(
+                arg.name.clone(),
+                arg.binding.clone(),
+                pointer,
+                self.module.global_variables[arg.handle].ty,
+                &mut expressions,
+                &mut |name, pointer, ty, binding, expressions| {
+                    members.push(StructMember {
+                        name,
+                        ty,
+                        binding: Some(binding),
+                        offset: span,
+                    });
+
+                    span += self.module.types[ty].inner.span(&self.module.constants);
+
+                    let len = expressions.len();
+                    let load = expressions.append(Expression::Load { pointer }, Default::default());
+                    body.push(
+                        Statement::Emit(expressions.range_from(len)),
+                        Default::default(),
+                    );
+                    components.push(load)
+                },
+            )
         }
 
         let (ty, value) = if !components.is_empty() {
             let ty = self.module.types.insert(
                 Type {
                     name: None,
-                    inner: TypeInner::Struct {
-                        top_level: false,
-                        members,
-                        span,
-                    },
+                    inner: TypeInner::Struct { members, span },
                 },
                 Default::default(),
             );
