@@ -99,6 +99,12 @@ pub enum ExpressionError {
     InvalidDepthReference(Handle<crate::Expression>),
     #[error("Depth sample level can only be Auto or Zero")]
     InvalidDepthSampleLevel,
+    #[error("Gather level can only be Zero")]
+    InvalidGatherLevel,
+    #[error("Gather component {0:?} doesn't exist in the image")]
+    InvalidGatherComponent(crate::SwizzleComponent),
+    #[error("Gather can't be done for image dimension {0:?}")]
+    InvalidGatherDimension(crate::ImageDimension),
     #[error("Sample level (exact) type {0:?} is not a scalar float")]
     InvalidSampleLevelExactType(Handle<crate::Expression>),
     #[error("Sample level (bias) type {0:?} is not a scalar float")]
@@ -343,6 +349,7 @@ impl super::Validator {
             E::ImageSample {
                 image,
                 sampler,
+                gather,
                 coordinate,
                 array_index,
                 offset,
@@ -466,6 +473,26 @@ impl super::Validator {
                     }
                 }
 
+                if let Some(component) = gather {
+                    match dim {
+                        crate::ImageDimension::D2 | crate::ImageDimension::Cube => {}
+                        crate::ImageDimension::D1 | crate::ImageDimension::D3 => {
+                            return Err(ExpressionError::InvalidGatherDimension(dim))
+                        }
+                    };
+                    let max_component = match class {
+                        crate::ImageClass::Depth { .. } => crate::SwizzleComponent::X,
+                        _ => crate::SwizzleComponent::W,
+                    };
+                    if component > max_component {
+                        return Err(ExpressionError::InvalidGatherComponent(component));
+                    }
+                    match level {
+                        crate::SampleLevel::Zero => {}
+                        _ => return Err(ExpressionError::InvalidGatherLevel),
+                    }
+                }
+
                 // check level properties
                 match level {
                     crate::SampleLevel::Auto => ShaderStages::FRAGMENT,
@@ -739,7 +766,17 @@ impl super::Validator {
                             false
                         }
                     },
-                    Bo::And | Bo::ExclusiveOr | Bo::InclusiveOr => match *left_inner {
+                    Bo::And | Bo::InclusiveOr => match *left_inner {
+                        Ti::Scalar { kind, .. } | Ti::Vector { kind, .. } => match kind {
+                            Sk::Bool | Sk::Sint | Sk::Uint => left_inner == right_inner,
+                            Sk::Float => false,
+                        },
+                        ref other => {
+                            log::error!("Op {:?} left type {:?}", op, other);
+                            false
+                        }
+                    },
+                    Bo::ExclusiveOr => match *left_inner {
                         Ti::Scalar { kind, .. } | Ti::Vector { kind, .. } => match kind {
                             Sk::Sint | Sk::Uint => left_inner == right_inner,
                             Sk::Bool | Sk::Float => false,
@@ -1321,12 +1358,20 @@ impl super::Validator {
                 }
                 ShaderStages::all()
             }
-            E::As { kind, convert, .. } => {
-                match convert {
-                    Some(width) if !self.check_width(kind, width) => {
-                        return Err(ExpressionError::InvalidCastArgument)
-                    }
-                    _ => {}
+            E::As {
+                expr,
+                kind,
+                convert,
+            } => {
+                let base_width = match *resolver.resolve(expr)? {
+                    crate::TypeInner::Scalar { width, .. }
+                    | crate::TypeInner::Vector { width, .. }
+                    | crate::TypeInner::Matrix { width, .. } => width,
+                    _ => return Err(ExpressionError::InvalidCastArgument),
+                };
+                let width = convert.unwrap_or(base_width);
+                if !self.check_width(kind, width) {
+                    return Err(ExpressionError::InvalidCastArgument);
                 }
                 ShaderStages::all()
             }
